from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

try:
    getattr(settings, "LOGGING_APP_NAME", None)
except ImproperlyConfigured:
    settings.configure()


LOGGING_APP_NAME = getattr(settings, "LOGGING_APP_NAME", "Default")
assert LOGGING_APP_NAME, "LOGGING_APP_NAME settings parameter must be defined"

LOGGING_ENCODING = getattr(settings, "LOGGING_ENCODING", "utf-8")
LOGGING_SERIALIZER = getattr(settings, "LOGGING_SERIALIZER", "json")
LOGGING_DATETIME_FORMAT = getattr(settings, "LOGGING_DATETIME_FORMAT", "%Y-%m-%d %H:%M:%S.%f")
LOGGING_USER_PROPERTY_NAME = getattr(settings, "LOGGING_USER_PROPERTY_NAME", "user")
LOGGING_FIELDS = getattr(
    settings, "LOGGING_FIELDS", ("levelname", "name", "module", "process", "thread", "pathname", "created")
)
DEVELOP = getattr(settings, "DEVELOP", False)

# Orjson options
LOGGING_OPT_NON_STR_KEYS = getattr(settings, "LOGGING_OPT_NON_STR_KEYS", False)
LOGGING_OPT_INDENT_2 = getattr(settings, "LOGGING_OPT_INDENT_2", False)
LOGGING_OPT_APPEND_NEWLINE = getattr(settings, "LOGGING_OPT_APPEND_NEWLINE", False)
LOGGING_OPT_NAIVE_UTC = getattr(settings, "LOGGING_OPT_NAIVE_UTC", False)
LOGGING_OPT_OMIT_MICROSECONDS = getattr(settings, "LOGGING_OPT_OMIT_MICROSECONDS", False)
LOGGING_OPT_PASSTHROUGH_DATACLASS = getattr(settings, "LOGGING_OPT_PASSTHROUGH_DATACLASS", False)
LOGGING_OPT_PASSTHROUGH_DATETIME = getattr(settings, "LOGGING_OPT_PASSTHROUGH_DATETIME", False)
LOGGING_OPT_SERIALIZE_DATACLASS = getattr(settings, "LOGGING_OPT_SERIALIZE_DATACLASS", False)
LOGGING_OPT_SERIALIZE_NUMPY = getattr(settings, "LOGGING_OPT_SERIALIZE_NUMPY", False)
LOGGING_OPT_SERIALIZE_UUID = getattr(settings, "LOGGING_OPT_SERIALIZE_UUID", False)
LOGGING_OPT_SORT_KEYS = getattr(settings, "LOGGING_OPT_SORT_KEYS", False)
LOGGING_OPT_STRICT_INTEGER = getattr(settings, "LOGGING_OPT_STRICT_INTEGER", False)
LOGGING_OPT_UTC_Z = getattr(settings, "LOGGING_OPT_UTC_Z", False)



from django.utils.log import DEFAULT_LOGGING

SECRET_KEY = "fake-key"
INSTALLED_APPS = [
    "django_json_logging",
    "tests",
]

LOGGING_APP_NAME = "test-app"
LOGGING_LOGLEVEL = "INFO"
LOGGING_FIELDS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
}

LOGGING_SERIALIZER = "orjson"
LOGGING_OPT_INDENT_2 = True
LOGGING_OPT_NON_STR_KEYS = True

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "json": {
            "()": "django_json_logging.formatters.JSONFormatter",
        },
        "django.server": DEFAULT_LOGGING["formatters"]["django.server"],
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "json",
        },
        "django.server": DEFAULT_LOGGING["handlers"]["django.server"],
    },
    "loggers": {
        "": {
            "handlers": ["console"],
            "level": LOGGING_LOGLEVEL,
        },
        "celery": {
            "handlers": ["console"],
            "level": LOGGING_LOGLEVEL,
        },
        "django": {
            "handlers": ["console"],
            "level": LOGGING_LOGLEVEL,
            "propagate": False,
        },
        "django.server": DEFAULT_LOGGING["loggers"]["django.server"],
    },
}



import json
from functools import lru_cache, partial
from typing import Callable, Union

from django_json_logging import settings

try:
    import ujson
except ImportError:  # pragma: nocover
    ujson = None  # type: ignore

try:
    import orjson
except ImportError:  # pragma: nocover
    orjson = None  # type: ignore


class ORJsonSerializer:
    options = {
        "LOGGING_OPT_INDENT_2": orjson.OPT_INDENT_2,
        "LOGGING_OPT_NON_STR_KEYS": orjson.OPT_NON_STR_KEYS,
        "LOGGING_OPT_APPEND_NEWLINE": orjson.OPT_APPEND_NEWLINE,
        "LOGGING_OPT_NAIVE_UTC": orjson.OPT_NAIVE_UTC,
        "LOGGING_OPT_OMIT_MICROSECONDS": orjson.OPT_OMIT_MICROSECONDS,
        "LOGGING_OPT_PASSTHROUGH_DATACLASS": orjson.OPT_PASSTHROUGH_DATACLASS,
        "LOGGING_OPT_PASSTHROUGH_DATETIME": orjson.OPT_PASSTHROUGH_DATETIME,
        "LOGGING_OPT_SERIALIZE_DATACLASS": orjson.OPT_SERIALIZE_DATACLASS,
        "LOGGING_OPT_SERIALIZE_NUMPY": orjson.OPT_SERIALIZE_NUMPY,
        "LOGGING_OPT_SERIALIZE_UUID": orjson.OPT_SERIALIZE_UUID,
        "LOGGING_OPT_SORT_KEYS": orjson.OPT_SORT_KEYS,
        "LOGGING_OPT_STRICT_INTEGER": orjson.OPT_STRICT_INTEGER,
        "LOGGING_OPT_UTC_Z": orjson.OPT_UTC_Z,
    }

    @staticmethod
    @lru_cache()
    def dumps() -> Callable:
        option = False
        for key, val in ORJsonSerializer.options.items():
            if getattr(settings, key):
                option = option | ORJsonSerializer.options[key]
        return partial(orjson.dumps, option=option) if option else orjson.dumps

    @classmethod
    def to_json(cls, dict_record: dict) -> str:
        """Converts record dict to a JSON string.

        The library orjson returns a bytes not an str.
        """
        assert orjson is not None, "orjson must be installed to use ORJsonSerializer"

        return str(cls.dumps()(dict_record), settings.LOGGING_ENCODING)


class UJsonSerializer:
    @staticmethod
    def to_json(dict_record: dict) -> str:
        """Converts record dict to a JSON string.

        The library ujson returns a bytes not an str.
        """
        assert ujson is not None, "ujson must be installed to use UJsonSerializer"

        if settings.DEVELOP:
            json_string = ujson.dumps(dict_record, ensure_ascii=False, indent=2)
        else:
            json_string = ujson.dumps(dict_record, ensure_ascii=False)
        return str(json_string, settings.LOGGING_ENCODING)


class JsonSerializer:
    @staticmethod
    def to_json(dict_record: dict) -> str:
        """Converts record dict to a JSON string.

        Using standard json library.
        """

        if settings.DEVELOP:
            return json.dumps(dict_record, ensure_ascii=False, indent=2)
        else:
            return json.dumps(dict_record, ensure_ascii=False)


def get_serializer() -> Union[ORJsonSerializer, UJsonSerializer, JsonSerializer]:
    mapping = {
        "orjson": ORJsonSerializer,
        "ujson": UJsonSerializer,
        "json": JsonSerializer,
    }
    assert settings.LOGGING_SERIALIZER in mapping, 'LOGGING_SERIALIZER must be "orjson", "ujson" or "json"'
    return mapping[settings.LOGGING_SERIALIZER]



import traceback
from logging import Formatter, LogRecord

from django_json_logging import settings
from django_json_logging.serializers import get_serializer

RECORD_ATTRIBUTES = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "message",
    "module",
    "msecs",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
}


class JSONFormatter(Formatter):
    """JSON log formatter."""

    datefmt = settings.LOGGING_DATETIME_FORMAT
    serializer = get_serializer()

    @staticmethod
    def get_extra(record: LogRecord) -> dict:
        """Determines and separate extra fields."""
        extra_names = set(record.__dict__.keys()).difference(RECORD_ATTRIBUTES)
        return {name: getattr(record, name) for name in extra_names}

    def format(self, record: LogRecord) -> str:
        """Format the specified record as text."""
        if "asctime" in settings.LOGGING_FIELDS:
            record.asctime = self.formatTime(record, self.datefmt)

        message = record.getMessage()
        dict_record = self.to_dict(message, record)
        return self.serializer.to_json(dict_record)

    def to_dict(self, message: str, record: LogRecord) -> dict:
        """Log record JSON formatter.

        Particular fields can be configured in settings.LOGGING_FIELDS.
        Request field is excluded because it can't be json-encoded directly.
        """
        extra = self.get_extra(record)
        default = {"message": message, "app_name": settings.LOGGING_APP_NAME}
        builtin = {field: getattr(record, field) for field in settings.LOGGING_FIELDS}
        dict_record = {**extra, **builtin, **default}

        dict_record.pop("request", None)

        if record.exc_info:
            if settings.DEVELOP:
                traceback.print_exception(*record.exc_info)
            else:
                extra["exc_info"] = self.formatException(record.exc_info)

        return dict_record
